# $Id: record.py 4403 2011-06-27 21:24:25Z place $
# $URL: https://svn.non-gnu.uvt.nl/uvt-dev/tags/sources/meresco/bzv-9.9.5/src/record.py $
from bzv.simplejson.encoder import JSONEncoder
from cgi import parse_qs
from cqlparser.cqlparser import parseString as parseCQL
from lxml.etree import parse, tostring
from meresco.components.http import utils as httputils
from meresco.core import Observable
from storage.hierarchicalstorage import HierarchicalStorageError
from traceback import format_exc
from urlparse import urlsplit
from xml.sax.saxutils import escape
from urllib import urlopen
import re

def okJSON():
    CRLF = "\r\n"
    ContentTypeJSON = "Content-Type: application/json"
    Ok = "HTTP/1.0 200 Ok" + CRLF
    okJSON = "HTTP/1.0 200 OK" + CRLF + \
             ContentTypeJSON + CRLF + \
             CRLF
    return okJSON

def forbiddenXml():
    return "HTTP/1.0 403 Forbidden" + httputils.CRLF + \
            "Content-Type: " + httputils.ContentTypeXml  + httputils.CRLF + \
        httputils.CRLF

def add(aDictionary, key, value):
    if value:
            aDictionary[key] = value

def makeUrlContent(url):
    s = '<a target="_blank" href="%(url)s">%(url)s</a>' %{'url': escape(url)}
    return s

def makeDoiContent(doi):
    s = '<a target="_blank" href="http://dx.doi.org/%(doi)s">%(doi)s</a>' %{'doi': doi}
    return s

def makePermaLink(id):
    s = '<a target="_blank" href="/id/%(id)s">%(id)s</a>' %{'id': escape(id)}
    return s

class Record(Observable):
    def __init__(self, infra = None):
         Observable.__init__(self)
         self.infra = infra
         self.infra.logging.debug('Starting record logging')
         self.logger = infra.logging.getLogger('meresco.record')
         self.logger.debug('Starting record logger')
         self.serverConfig  = self.infra.configRoot
         self.languageDict = self.serverConfig.getSectionNormalized('languages')
         self.genreDict = self.serverConfig.getSectionNormalized('genre')
         self.nsMap = self.serverConfig.getSectionDict('namespaces')

    def createRecord(self, recordId, format = "full"):
        record = _Record(recordId, self.any, self.logger, self.nsMap, self.serverConfig, self.languageDict, self.genreDict)

        if format == "full":
            record.makeDictionary()
        elif format == "short":
            record.makeShortDict()
        elif format == "citec":
            record.makeCitecDict()
        elif format == "availability":
        #FIXME niet geimplementeerd
            record.makeAvailDict()
        return record.record

    def getMods(self, recordId):
        try:
            _mods = parse(self.any.getStream(recordId, 'mods'))
        except HierarchicalStorageError:
            url = baseurl + \
                "/sru?version=1.1&operation=searchRetrieve&recordSchema=mods&query=meta.upload.id+exact+\"" + \
                recordId + "\""
            _mods = parse(urlopen(url))
            self.logger.debug("mods from external system: " + url)
        return _mods

    def getSid(self, recordId):
        try:
            _meta = parse(self.any.getStream(recordId, 'meta'))
        except HierarchicalStorageError:
            url = baseurl + \
                "/sru?version=1.1&operation=searchRetrieve&recordSchema=meta&query=meta.upload.id+exact+\"" + \
                recordId + "\""
            _meta = parse(urlopen(url))
            self.logger.debug("meta from external system: " + url)
        self._collection = _meta.xpath('//meta:meta/meta:repository/meta:collection/text()', self.nsMap)[0]
        self._db         = _meta.xpath('/meta:meta/meta:repository/meta:repositoryGroupId/text()', self.nsMap)[0]
        return self._collection + ':' + self._db

    def getOrder(self):
###### Note that all fields must be listed here
######      otherwise they will not be displayed.
###### This is a public method.
       order = {
           1: 'Title',
           3: 'Author(s)',
           4: 'Collaborator(s)',
           5: 'Editor(s)',
           6: 'Supervisor(s)',
           7: 'Corporation',
           8: 'Source',
           9: 'Publication date',
           10: 'URL(s)',
           11: 'Doi',
           12: 'Dataset',
           13: 'Abstract',
           14: 'Person as subject',
           15: 'Keyword(s)',
           16: 'Classification',
           17: 'Generated JEL',
           18: 'ISSN',
           19: 'ISBN',
           20: 'Publisher',
           21: 'Extent',
           22: 'Series',
           23: 'Edition',
           24: 'Note(s)',
           25: 'Reference',
           26: 'Publication type',
           27: 'Location',
           28: 'Related Titles',
           29: 'Cited by',
           30: 'References',
           31: 'Co-citing',
           32: 'Language',
           33: 'RecordID',
           ## dataplus, evs
           ## nu opgelost met JSON en javascript
           #34: 'Operationalisations'
       }
       return order

class _Record(object):
    def __init__(self, recordId, any, logger, nsMap, serverConfig = None, languageDict = None, genreDict = None):
        self._recordId = recordId
        self.any = any
        self.logger = logger
        self.serverConfig = serverConfig
        self.languageDict = languageDict
        self.genreDict = genreDict
        self.nsMap = nsMap
        baseurl = self.serverConfig.get('url', section='dbserver')
        recId = recordId
        self.logger.debug(recId)
        try:
            if recordId.split(':')[0] in ['opc-uvt-nl']:
                try:
                    _metadata = parse(self.any.getStream(recId, 'mods'))
                    _meta     = parse(self.any.getStream(recId, 'meta'))
                except HierarchicalStorageError:
                    url = baseurl + \
                            "/sru?version=1.1&operation=searchRetrieve&recordSchema=mods&x-recordSchema=meta&query=meta.upload.id+exact+\"" + \
                            recordId + "\""
                    _metadata = parse(urlopen(url))
                    _meta = _metadata
                    self.logger.debug("metadata from external system: " + url)
            else:
                try:
                    _metadata = parse(self.any.getStream(recId, 'metadata'))
                    _meta     = parse(self.any.getStream(recId, 'meta'))
                except HierarchicalStorageError:
                    url = baseurl + "/sru?version=1.1&operation=searchRetrieve&x-recordSchema=meta&query=meta.upload.id+exact+\"" + recordId + "\""
                    self.logger.debug("Cannot retrieve record; remote host will be tried, url +'" + url +'"')
                    _metadata = parse(urlopen(url))
                    _meta = _metadata
                    self.logger.debug("metadata from external system: " + url)

            self._parts      = self.nodeset(_meta, '//meta:parts/meta:part')
            self._recordId   = recordId
            self._collection = self.xpath(_meta, '//meta:meta/meta:repository/meta:collection/text()')
            self._metadata = _metadata

            try:
                self._setGenre()
            except Exception:
                self._genre=''
                self.logger.error("no Genre for " + self._recordId)

        except Exception:
            self.logger.error("No metadata for record: " + str(recordId))
            self.logger.error(format_exc())

    def xpath(self, node, path):
        result = node.xpath(path, self.nsMap)
        return result and result[0] or ''

    def nodeset(self, node, path):
        return node.xpath(path, self.nsMap)

    def _setGenre(self):
        genre = self.xpath(self._metadata, '//mods:mods/mods:genre/text()').rstrip("/")
        genre = self.genreDict[genre]
        self._genre = genre

    def _addName(self, names, name):
        family = ""
        given  = ""
        infix  = ""
        for namePart in self.nodeset(name, 'mods:namePart'):
            type = self.xpath(namePart, '@type')
            if type == 'family' or not type:
                n = self.xpath(namePart, 'text()')
                f = n.split(',', 1)
                # Hippo, van
                if len(f) == 2:
                    family = f[0]
                    if type: infix = f[1]
                    else: given = f[1]
                # van Hippo
                # no comma
                else:
                    n = n.replace('  ', ' ')
                    if type: family = n
                    else:
                        f = n.split(' ')
                        family = f[-1]
                        if not type: given = ' '.join(f[0:-1])
            elif type == 'given':  given  = self.xpath(namePart, 'text()')
        if family: names += family
        if given:
            names  += ', ' + given
            if infix: names += ' ' + infix
        if family:
            dai_nl = "" # wordt niet gebruikt
            anr = ""
            ID = name.xpath("@ID")
            if ID:
                ID = ID[0]
                xpath = "//mods:mods/mods:extension/dai:daiList/dai:identifier[@IDref='%s']" %ID
                try:
                    DAIs = self._metadata.xpath(xpath, self.nsMap)
                    for dai in DAIs:
                        if dai.xpath("@authority") and dai.xpath("@authority")[0] == "info:eu-repo/dai/nl":
                            dai_nl = dai.text
                        elif dai.text.startswith("http://www.tilburguniversity.nl/dai/"):
                            anr = dai.text.split("/dai/")[1]
                except: pass
            if anr:
                url = escape("http://www.tilburguniversity.edu/webwijs/show/?anr=" + anr)
                names += ' (<a href="' + url + '">info</a>)'
            names += '; '
        return names

    def _citec(self):
        res = {}
        if self._collection == "repec":
            parts = self._recordId.split(':', 5)
            url = "ftp://ftp.repec.org/opt/amf/CitEc/" +  parts[3] + "/" + parts[4] + "/" + parts[5] + ".amf.xml"
            self.logger.debug("citec url: " + url)
            try:
                amf = parse(urlopen(url))
            except Exception:
                return {}
            refs = []
            references = amf.xpath("/amf:amf/amf:text/amf:references", namespaces=self.nsMap)
            if len(references) > 0:
                for r in references:
                    refid = r.xpath("amf:text/@ref", namespaces=self.nsMap)[0]
                    s = refid.split(":")
                    recId = "RePEc-"+s[1]+"-"+s[2]+":oai:"+refid
                    record = self.any.createRecord(recId, format = "short")
                    if not record:
                        recId = ""
                        refstring = r.xpath("acis:referencestring", namespaces=self.nsMap)[0].text
                        record = {"Citation": refstring}
                    refs.append((recId, record))
            references = amf.xpath("/amf:amf/amf:text/amf:reference", namespaces=self.nsMap)
            if len(references) > 0:
                for r in references:
                    string = r.xpath("amf:literal", namespaces=self.nsMap)[0].text
                    record = {"Citation": string}
                    refs.append(("", record))
            citedBy = []
            citedBys = amf.xpath("/amf:amf/amf:text/amf:isreferencedby", namespaces=self.nsMap)
            if len(citedBys) > 0:
                for c in citedBys:
                    by = c.xpath("amf:text/@ref", namespaces=self.nsMap)[0]
                    #RePEc:dgr:kubcen:200847
                    #RePEc-dgr-kubcen:oai:RePEc:dgr:kubcen:200847
                    s = by.split(":")
                    recId = "RePEc-"+s[1]+"-"+s[2]+":oai:"+by
                    record = self.any.createRecord(recId, format = "short")
                    citedBy.append((recId, record))
            if citedBy: res['Cited by']  = citedBy
            if refs:
                res['References'] = refs
                res['Co-citing'] = "http://citec.repec.org/cgi-bin/get_data.pl?o=all&h=repec:%s:%s:%s#cociting" %(parts[3], parts[4], parts[5])
        return res

    def _ddi(self):
        ddi = {}
        if self._parts:
            for part in self._parts:
                if part.text == "ddi":
                    dd = parse(self.any.getStream(self._recordId, 'ddi'))
                    ddi['URL'] = self.xpath(dd, '/ddi:codeBook/ddi:docDscr/ddi:citation/ddi:holdings/@URI')
                    accsPlac  = self.xpath(dd, '/ddi:codeBook/ddi:stdyDscr/ddi:dataAccs/ddi:setAvail/ddi:accsPlac/text()')
                    if accsPlac:
                        p = re.compile("<a href=[\'\"]([^\'\"]*)[\'\"]>([^<]*)")
                        m = p.match(accsPlac)
                        if m:
                            ddi['Files'] = {}
                            ddi['Files']['Dataset'] = {}
                            ddi['Files']['Dataset']['Data Access Place'] = {}
                            ddi['Files']['Dataset']['Data Access Place']['URI']  = m.group(1)
                            ddi['Files']['Dataset']['Data Access Place']['Text'] = m.group(2)
                    files = self.nodeset(dd, '/ddi:codeBook/ddi:otherMat[@level="study"]')
                    if files:
                        fileDict = {}
                        for file in files:
                            name     = self.xpath(file, 'ddi:labl/text()')
                            URI      = self.xpath(file, '@URI')
                            Text     = self.xpath(file, 'ddi:txt/text()')
                            Category = self.xpath(file, 'ddi:notes[@type="vdc:category"]/text()')
                            if Category not in fileDict: fileDict[Category] = {}
                            fileDict[Category][name] = {}
                            fileDict[Category][name]['URI']  = URI
                            fileDict[Category][name]['Text'] = Text
                        ddi['Files'] = fileDict
        return ddi

    def _ddi3(self):
        ddi3 = {}
        if self._parts:
            for part in self._parts:
                if part.text == "ddi3":
                    dd = parse(self.any.getStream(self._recordId, 'ddi3'))
                    concepts = self.nodeset(dd, '/ddi3:DDIInstance/s:StudyUnit/l:LogicalProduct/l:VariableScheme/l:VariableGroup')
                    if concepts:
                        conceptDict = {}
                        for concept in concepts:
                            name = self.xpath(concept, 'r:Label/text()')
                            variables = []
                            conceptDict[name] = {}
                            variableURNs = self.nodeset(concept, 'l:VariableReference/r:URN')
                            if variableURNs:
                                for URN in variableURNs:
                                    path = '/ddi3:DDIInstance/s:StudyUnit/l:LogicalProduct/l:VariableScheme/l:Variable[@urn="%s"]' %URN.text
                                    var = self.nodeset(dd, path)[0]
                                    varLabel = self.xpath(var, 'r:Label/text()')
                                    varDescription = self.xpath(var, 'r:Description/text()')
                                    variables.append((varLabel, varDescription))
                            conceptDict[name]['variables'] = variables
                            syntax = self.xpath(concept, 'r:Description/text()')
                            if syntax: conceptDict[name]['syntax'] = syntax
                        ddi3['Concepts'] = conceptDict
                    syntax = self.xpath(dd, "/ddi3:DDIInstance/s:StudyUnit/r:Note/r:Content/text()")
                    if syntax: ddi3['Syntax'] = syntax
                    universe = dd.xpath("/ddi3:DDIInstance/s:StudyUnit/c:ConceptualComponent/c:UniverseScheme/c:Universe/c:SubUniverse/c:HumanReadable/text()", self.nsMap)
                    if universe:
                        ddi3["Universe"] = {}
                        for subUniverse in universe:
                            countryWave = subUniverse.split(' -- ')
                            country = countryWave[0]
                            wave = countryWave[1]
                            if not wave in ddi3["Universe"]: ddi3["Universe"][wave]= []
                            ddi3["Universe"][wave].append(country)
        return ddi3

    def _getDataset(self):
        if self._parts:
            for part in self._parts:
                if part.text == "ddi":
                    ddi = parse(self.any.getStream(self._recordId, 'ddi'))
                    url = ddi.xpath("//ddi:codeBook/ddi:docDscr/ddi:citation/ddi:holdings/@URI", namespaces=self.nsMap)[0]
                    return url
        return ''

    def _getDate(self):
        date = self.xpath(self._metadata, '//mods:originInfo/mods:dateIssued/text()')
        if date:
            altDate = date.split('(', 1)
            # bij catalogusrecords staat de goede datum soms tussen haakjes
            if len(altDate) == 2: date = altDate[1].rstrip(')')
            else:
                range = date.split('-', 1)
                if len(range) == 2:
                    # echte range yyyy-yyyy
                    if re.match(r'\d{4}', range[1]): pass
                    # yyyy-mm(-dd) is geen echte range
                    else: date = range[0]
            return date

    def _getEdition(self):
        return self.xpath(self._metadata, '//mods:originInfo/mods:edition/text()')
    
    def _getExtent(self):
        return self.xpath(self._metadata, '//mods:mods/mods:physicalDescription/mods:extent/text()')

    def _getKeywords(self):
        keywords = ""
        for topic in self.nodeset(self._metadata, '//mods:mods/mods:subject/mods:topic'):
            keywords += self.xpath(topic, 'text()') + '; '
        if keywords: return keywords.rstrip('; ')
        else: return ''

    def _getLanguage(self):
        languages = ""
        for languageTerm in self.nodeset(self._metadata, '//mods:mods/mods:language/mods:languageTerm'):
            language = self.xpath(languageTerm, 'text()')
            try: 
                language = self.languageDict[language]
                language = self.serverConfig.get(language, section = 'languages.translations', language='en')
            except: pass
            languages += language + '; '
        if languages: return languages.rstrip('; ')
        else: return ''

    def _getLocation(self):
        locations = self.nodeset(self._metadata, '//mods:mods/mods:location/mods:holdingSimple/mods:copyInformation/mods:shelfLocator')
        if locations:
            parts = self._recordId.split(':')
            s = []
            for loc in locations:
                titles = []
                # varianten:
                # u"Zie afzonderl\u0133ke publikaties"
                # "Zie afzonderlijke publicaties"
                # "zie afzonderlijke publicaties"
                # "zie afzonderlijke publ."
                #if loc.text.find("ie afzonderl") > 0:
                signatuur =  self.xpath(loc, "./text()")
                if signatuur.find("ie afzonderl") > 0:
                    query = "mods.relatedItem.recordInfo.recordIdentifier=" + parts[1]
                    stop = 100
                    total, recordIds = self._search(query, stop)
                    if total > 0 and total < stop + 1:
                        for recordId in recordIds:
                            record = self.any.createRecord(recordId, format='short')
                            url = "/id/" + recordId
                            titles.append((url, record['Title']))
                    elif total > stop:
                        url = "?query=%s+%s" %(parts[0], parts[1])
                        titles.append((url, "Display the separate parts"))
                    else:
                        titles.append(("", "No publications found"))

                holdingset = self.nodeset(loc, "../mods:enumerationAndChronology")
                holdings = []
                for holding in holdingset:
                    holdings.append(holding.text)

                noteset = self.nodeset(loc, "../mods:note")
                notes = []
                for note in noteset:
                    if not note.get("type"): notes.append(note.text)
                
                s.append((signatuur, titles, notes, holdings))
            return s

    def _search(self, query, stop):
        parsetree = parseCQL(query)
        return self.any.executeCQL(cqlAbstractSyntaxTree=parsetree, start=0, stop=stop)
        # sortBy=sortBy, sortDescending=sortDescending 

    def _getPersonAsSubject(self):
        names = ""
        for name in self.nodeset(self._metadata, '//mods:mods/mods:subject/mods:name/mods:namePart'):
            names += name.text + '; '
        if names: return names.rstrip('; ')
        else: return ''

    def _getRelatedTitles(self):
        pass

    def _getSeries(self):
        out = ""
        series = (self.nodeset(self._metadata, '//mods:relatedItem[@type="series"]'))
        if series:
            for aseries in series:
                title = escape(self.xpath(aseries, 'mods:titleInfo/mods:title/text()'))
                if not title: continue
                ppn   = self.xpath(aseries, 'mods:recordInfo/mods:recordIdentifier/text()')
                part  = self.xpath(aseries, 'mods:part/mods:detail/mods:number/text()')
                if ppn:
                    out += '<a target="_blank" href="/?query=%s+%s">%s</a>' %(self._recordId.split(':')[0], ppn, title)
                else:
                    out += title
                if part: out += '<i>: ' + escape(part) + '</i>'
                out += '; '
            return out.rstrip('; ')

    def _getSource(self):
        editors = escape(self._getHostEditors())
        title = escape(self._getHostTitle())
        if not title: return ''
        if editors: return editors + ' ' + '<i>' + title + '</i>'
        date = self._getDate()
        volume    = self.xpath(self._metadata, '//mods:part/mods:detail[@type="volume"]/mods:number/text()')
        issue     = self.xpath(self._metadata, '//mods:part/mods:detail[@type="issue"]/mods:number/text()')
        firstpage = self.xpath(self._metadata, '//mods:part/mods:extent/mods:start/text()')
        endpage   = self.xpath(self._metadata, '//mods:part/mods:extent/mods:end/text()')
        if volume: title += ' vol. ' + volume
        if date: title += ' (' + date + ')'
        if issue: title += ' nr. ' + issue
        if firstpage:
            title += ' p.' + firstpage
            if endpage: title += '-' + endpage
        part = self.xpath(self._metadata, '//mods:relatedItem[@type="host"]/mods:part/mods:text/text()')
        if part: title += '. ' + part
        return title

    def _getTitle(self):
        title = self.xpath(self._metadata, '//mods:mods/mods:titleInfo/mods:title/text()')
        if not title:
            title = self.xpath(self._metadata, '//mods:relatedItem[@type="host"]/mods:titleInfo/mods:title/text()')
            if title: title = '[' + title + ']'
            #fixme: naar config
            else: title = '[NO TITLE AVAILABLE]'
        subTitles = self.nodeset(self._metadata, '//mods:mods/mods:titleInfo/mods:subTitle/text()')
        if subTitles:
            for subTitle in subTitles:
                title += ": " + subTitle
        return escape(title)

    def _handleNames(self):
        authors = ""
        collaborators = ""
        editors = ""
        supervisors = ""
        corporation = ""
        modsNames = (self.nodeset(self._metadata, '//mods:mods/mods:name'))
        if not modsNames: return {}
        for name in modsNames:
            role = self.xpath(name, 'mods:role/mods:roleTerm/text()').strip()
            if   role == 'aut' or not role:
                authors       = self._addName(authors,       name)
            elif role == 'clb':
                collaborators = self._addName(collaborators, name)
            elif role == 'edt':
                editors       = self._addName(editors,       name)
            elif role == 'ths':
                supervisors   = self._addName(supervisors,   name)
            elif role == 'dgg':
                content = 'Degree granted by ' + str(self.xpath(name, 'mods:namePart/text()'))
                corporation   = content
        names = {}
        if corporation:
            names['Corporation'] = corporation
        if authors:
            names['Author(s)'] = authors.rstrip(' ;')
        if collaborators:
            names['Collaborator(s)'] = collaborators.rstrip(' ;')
        if editors:
            names['Editor(s)'] = editors.rstrip(' ;')
        if supervisors:
            names['Supervisor(s)'] = supervisors.rstrip(' ;')
        return names

    def _getHostTitle(self):
        return self.xpath(self._metadata, '//mods:mods/mods:relatedItem[@type="host"]/mods:titleInfo/mods:title/text()')

    def _getHostEditors(self):
        editors = ""
        family = ""
        given = ""
        for name in self.nodeset(self._metadata, '//mods:mods/mods:relatedItem/mods:name'):
            if self.xpath(name, 'mods:role/mods:roleTerm/text()') == 'edt':
                for namePart in self.nodeset(name, 'mods:namePart'):
                    type = self.xpath(namePart, '@type')
                    if type == 'family':  family = self.xpath(namePart, 'text()')
                    elif type == 'given':  given = self.xpath(namePart, 'text()')
                if family: editors += family
                if given:  editors += ', ' + given
                if family: editors += '; '
                family = ""
                given = ""
        if editors: return editors.rstrip('; ') + ' (Ed.)'
        else: return ''

    def _getAbstract(self):
        return escape(self.xpath(self._metadata, '//mods:abstract/text()'))

    def _getUrls(self):
        urls = []
        for url in self.nodeset(self._metadata, '//mods:location/mods:url'):
            urls.append(url.text)
        for url in self.nodeset(self._metadata, '//mods:location/mods:holdingSimple/mods:copyInformation/mods:electronicLocator'):
            urls.append(url.text)
        for Item in self.nodeset(self._metadata, '//didl:DIDL/didl:Item/didl:Item'):
            if self.xpath(Item, 'didl:Descriptor/didl:Statement/rdf:type/text()')        == 'info:eu-repo/semantics/objectFile' or \
               self.xpath(Item, 'didl:Descriptor/didl:Statement/dip:ObjectType/text()')  == 'info:eu-repo/semantics/objectFile' or \
               self.xpath(Item, 'didl:Descriptor/didl:Statement/rdf:type/@rdf:resource') == 'info:eu-repo/semantics/objectFile':
                url = self.xpath(Item, 'didl:Component/didl:Resource/@ref')
                if url: urls.append(url)
        return urls

    def _getNotes(self):
        notes = [] 
        count = 0
        for note in self.nodeset(self._metadata, '//mods:mods/mods:note/text()'):
             notes.append(escape(note))
        return notes

    def _getReference(self):
        return self.xpath(self._metadata, '//mods:mods/mods:relatedItem[@type="isReferencedBy"]/mods:titleInfo/mods:title/text()')

    def _getPublisher(self):
        place = self.xpath(self._metadata, '//mods:mods/mods:originInfo/mods:place/mods:placeTerm/text()')
        name = self.xpath(self._metadata, '//mods:mods/mods:originInfo/mods:publisher/text()')
        publisher = ""
        if name:
           publisher = ""
           if place: publisher = place + ": "
           publisher += name
        return escape(publisher)

    def _getClassification(self):
        string = ""
        for classification in self.nodeset(self._metadata, '//mods:mods/mods:classification'):
            string += self.xpath(classification, 'text()') + '; '
        return string.rstrip('; ')

    def _getGeneratedJel(self):
        if self._parts:
            for part in self._parts:
                if part.text == "jel":
                    mods = parse(self.any.getStream(self._recordId, 'jel'))
                    string = ""
                    for classification in self.nodeset(mods, '//mods:mods/mods:classification'):
                        string += self.xpath(classification, 'text()') + '; '
                    return string.rstrip('; ')
        return ''

    def _handleIdentifiers(self):
        identifiers = (self.nodeset(self._metadata, '//mods:identifier'))
        res = {}

        if not identifiers: return res
        issns = ""
        isbns = ""
        doi   = ""
        for id in identifiers:
            idParts = self.xpath(id, 'text()').split(':')
            ### urn:{issn|issbn}:<value>
            ### info:{doi}/<value>
            if idParts[0] == 'urn' or idParts[0] == 'URN':
                if len(idParts) < 3: break
                if idParts[1] == 'issn' or idParts[1] == 'ISSN':
                    if issns: issns += "; "
                    issns += idParts[2]
                elif idParts[1] == 'isbn' or idParts[1] == 'ISBN':
                    if isbns: isbns += "; "
                    displayLabel = self.xpath(id, '@displayLabel')
                    if displayLabel: isbns += displayLabel
                    else: isbns += idParts[2]
            elif idParts[0] == 'info':
                parts = idParts[1].split('/', 1)
                if parts[0] == 'doi': doi = parts[1]

        if issns: res['ISSN'] = issns
        if isbns: res['ISBN'] = isbns
        if doi  : res['Doi']  = doi
        return res

    def makeDictionary(self):
        self.makeShortDict()
        add(self.record, 'Language', self._getLanguage())
        add(self.record, 'Abstract', self._getAbstract())
        add(self.record, 'Person as subject', self._getPersonAsSubject())
        add(self.record, 'Keyword(s)', self._getKeywords())
        add(self.record, 'Classification', self._getClassification())
        add(self.record, 'Generated JEL', self._getGeneratedJel())
        add(self.record, 'Publisher', self._getPublisher())
        add(self.record, 'Edition', self._getEdition())
        add(self.record, 'Extent', self._getExtent())
        add(self.record, 'Series', self._getSeries())
        add(self.record, 'Note(s)', self._getNotes())
        #add(self.record, 'Reference', self._getReference())
        add(self.record, 'Location', self._getLocation())
        add(self.record, 'Related Titles', self._getRelatedTitles())
        add(self.record, 'RecordID', self._recordId)
        #self.record.update(self._citec())
        add(self.record, 'Operationalisations', self._ddi3())
        add(self.record, 'Data files', self._ddi())

    def makeShortDict(self):
        self.record = {}
        add(self.record, 'Title', self._getTitle())
        self.record.update(self._handleNames())
        add(self.record, 'Source', self._getSource())
        add(self.record, 'Publication date', self._getDate())
        self.record.update(self._handleIdentifiers())
        add(self.record, 'Publication type', self._genre)
        add(self.record, 'URL(s)', self._getUrls())
        add(self.record, 'Dataset', self._getDataset())

    def makeCitecDict(self):
        self.record.update(self._citec())

class RecordXml(Observable):
    def __init__(self):
         Observable.__init__(self)

    def handleRequest(self, RequestURI='', **kwargs):
####### TP: toevoegen het afvangen van exceptions: 1. er is geen record id,
####### 2. er wordt geen metadata gevonden
        Scheme, Netloc, Path, Query, Fragment = urlsplit(RequestURI)
        arguments = parse_qs(Query)
        recordId = arguments.get('recordId', [''])[0]
        context  = arguments.get('context', [''])[0]
        if not self.any.recordAccessAllowed(recordId, kwargs):
            yield forbiddenXml()
            yield """<?xml version="1.0" encoding="UTF-8"?>
<denied/>"""
            return
        record = self.any.createRecord(recordId)
        yield httputils.okXml
        yield """<?xml version="1.0" encoding="UTF-8"?>
<record>"""

####### TP: labels zijn in het Engels. Toevoegen: ondersteunen van andere talen.
####### TP: ook handig is om per field het type toe te voegen (taalonafhankelijk)
        order = self.any.getOrder()
        order.keys().sort()
        for key in order.keys():
            label = order[key]
            if label in record:
                if label == 'URL(s)':
                    for url in record['URL(s)']:
                        yield '\n <field><label>URL</label><content>%s</content></field>' %(makeUrlContent(url))
                elif label == 'Doi':
                        yield '\n <field><label>Doi</label><content>%s</content></field>' %(makeDoiContent(record['Doi']))
                elif label == 'Dataset':
                        yield '\n <field><label>Dataset</label><content>%s</content></field>' %(makeUrlContent(record['Dataset']))
                elif label in ['Related Titles', 'Series', 'Source']:
                    yield '\n <field><label>%s</label><content>%s</content></field>' %(label, record[label])
                elif label == "Keyword(s)":
                    if context == "permalink": content = record[label]
                    else:
                        keywords = record["Keyword(s)"].split("; ")
                        content = ""
                        for keyword in keywords:
                            content += "<a href='search?query=mods.subject.topic=" + '"' + keyword.replace("'", " ") + '"' + "'>" + keyword + '</a>; '
                    yield '\n <field><label>%s</label><content>%s</content></field>' %(label, content.rstrip('; '))
                elif label in ['Location']:
                    yield '\n <field><label>%s</label><content>' %(label)
                    # items in record['Location']
                    # 0 = signatuur
                    # 1 = deeltitels
                    # 2 = notes
                    # 3 = holdings
                    for item in record[label]:
                        if item is not record[label][0]:
                            yield "<br/><br/>"
                        yield item[0]
                        if item[3]:
                            for holding in item[3]:
                                yield "<br/>" + holding
                        for note in item[2]:
                            yield  "<br/>" + note        
                        if item[1]:
                            titles = item[1]
                            yield '\n  <ul>'
                            for title in titles:
                                if title[0]:
                                    yield '\n   <li><a target="_blank" href="%s">%s</a></li>' %(title[0], escape(title[1]))
                                else:
                                    yield '\n   %s' %escape(title[1])
                            yield '\n  </ul>'
                    yield '</content></field>'
                elif label in ['Note(s)']:
                    yield '\n <field><label>%s</label><content>' %(label)
                    count = 0
                    for note in record[label]:
                        count += 1
                        if count > 1:
                            yield "<br/>"
                        yield note
                    yield '</content></field>'
                elif label in ['Cited by', 'References']:
                    yield '\n <field><label>%s</label><content><ul>' %(label)
                    for item in record[label]:
                        if item[0] and item[1]:
                            rec = item[1]
                            citation =  '<li>'
                            if 'Author(s)' in rec: citation += rec['Author(s)']
                            if 'Publication date' in rec:
                                citation += ' (' + \
                                            rec['Publication date'] + \
                                            ')'
                            citation += ' <a target="_blank" href="?query=' + \
                                        item[0] + \
                                        '&amp;displayType=single">' + \
                                        escape(rec['Title']) + \
                                        '</a> '
                            if 'Source' in rec: citation += rec['Source']
                            citation += '</li>'
                            yield '\n %s' %(citation)
                        elif item[1]:
                            yield '\n <li>%s</li>' %(escape(item[1]['Citation']))
                    yield '\n </ul></content></field>'
                elif label in ['Co-citing']:
                    yield '\n <field><label>Co-citing</label><content><a target="_blank" href="%s">Documents</a> which have cited the same bibliography</content></field>' %escape(record[label])
                elif label in ['Author(s)', 'Collaborator(s)', 'Editor(s)', 'Supervisor(s)']:
                    yield '\n <field><label>%s</label><content>%s</content></field>' %(label, record[label])
                elif label in ['RecordID']:
                    yield '\n <field><label>RecordID</label><content>%s</content></field>' %(makePermaLink(recordId))

                elif label == "Operationalisations":
                    pass
                    ## nu via JSON en javascript
                else:
                    yield '\n <field><label>%s</label><content>%s</content></field>' %(label, escape(record[label]))
        yield '\n</record>'

class Citec(Observable):
    def __init__(self):
         Observable.__init__(self)

    def handleRequest(self, RequestURI='', **kwargs):
####### TP: toevoegen het afvangen van exceptions: 1. er is geen record id,
####### 2. er wordt geen metadata gevonden
        Scheme, Netloc, Path, Query, Fragment = urlsplit(RequestURI)
        arguments = parse_qs(Query)
        recordId = arguments.get('recordId', [''])[0]
        record = self.any.createRecord(recordId, format="citec")
        yield httputils.okXml
        yield """<?xml version="1.0" encoding="UTF-8"?>
<record>"""
        for label in record.keys():
            if label in ['Cited by', 'References']:
                yield '\n <field><label>%s</label><content><ul>' %(label)
                for item in record[label]:
                    if item[0] and item[1]:
                        rec = item[1]
                        citation =  '<li>'
                        if 'Author(s)' in rec: citation += rec['Author(s)']
                        if 'Publication date' in rec:
                            citation += ' (' + \
                                        rec['Publication date'] + \
                                        ')'
                        citation += ' <a target="_blank" href="?query=' + \
                                    item[0] + \
                                    '&amp;displayType=single">' + \
                                    escape(rec['Title']) + \
                                    '</a> '
                        if 'Source' in rec: citation += rec['Source']
                        citation += '</li>'
                        yield '\n %s' %(citation)
                    elif item[1]:
                        yield '\n <li>%s</li>' %(escape(item[1]['Citation']))
                yield '\n </ul></content></field>'
            elif label in ['Co-citing']:
                yield '\n <field><label>Co-citing</label><content><a target="_blank" href="%s">Documents</a> which have cited the same bibliography</content></field>' %escape(record[label])
        yield '\n</record>'

class RecordJSON(Observable):
    def __init__(self):
         Observable.__init__(self)

    def handleRequest(self, RequestURI='', **kwargs):
####### TP: toevoegen het afvangen van exceptions: 1. er is geen record id,
####### 2. er wordt geen metadata gevonden
        Scheme, Netloc, Path, Query, Fragment = urlsplit(RequestURI)
        arguments = parse_qs(Query)
        recordId = arguments.get('recordId', [''])[0]
        record = self.any.createRecord(recordId)
        yield okJSON()
        yield JSONEncoder().encode(record)
